# In consumers.py
from channels import Group, Channel
from .models import ChatMessage, NicknameRegistration
from otree.models.participant import Participant
import json

def get_chat_group(channel):
    return 'otreechat-{}'.format(channel)

from channels.generic.websockets import JsonWebsocketConsumer

def msg_consumer(message):
    content = message.content

    # For now, don't create a model because we
    # don't yet have a way to export this table.
    # currently oTree admin is not extensible.

    channel = content['channel']
    grp = get_chat_group(channel)

    # list containing 1 element
    # TODO: JS needs to send participant code to server,
    # but server needs to send nickname to client


    participant_id = Participant.objects.filter(code=content['participant_code']).values_list(
        'id', flat=True)[0]

    nickname = NicknameRegistration.objects.values_list(
        'nickname', flat=True).get(participant=participant_id, channel=channel)

    chat_message = {
        'channel': channel,
        'nickname': nickname,
        'body': content['body'],
        'participant_id': participant_id
    }

    Group(grp).send({'text': json.dumps([chat_message])})

    ChatMessage.objects.create(
        participant_id=participant_id,
        channel=channel,
        body=content['body'],
        nickname=nickname
    )


class ChatConsumer(JsonWebsocketConsumer):

    # Set to True if you want it, else leave it out
    strict_ordering = False

    def connection_groups(self, **kwargs):
        """
        Called to return the list of groups to automatically add/remove
        this connection to/from.
        """
        return [get_chat_group(kwargs['channel'])]

    def connect(self, message, **kwargs):
        history = ChatMessage.objects.filter(
            channel=kwargs['channel']).order_by('timestamp').values(
                'channel', 'nickname', 'body', 'participant_id'
        )

        # Convert ValuesQuerySet to list
        self.send(list(history))

    def receive(self, content, **kwargs):
        # Stick the message onto the processing queue
        Channel("otree.chat_messages").send(content)



from channels.routing import route, route_class
from otreechat.consumers import msg_consumer, ChatConsumer
from otree.channels.routing import channel_routing

channel_routing += [
    route_class(ChatConsumer, path=r"^/otreechat/(?P<channel>[a-zA-Z0-9_-]+)/$"),
    route('otree.chat_messages', msg_consumer),
]



from django import template
register = template.Library()
import re
from otreechat.models import NicknameRegistration
import json
from django.db.models.query import QuerySet
from django.utils.safestring import mark_safe
from otree.api import safe_json

@register.inclusion_tag('otreechat/widget.html', takes_context=True)
def chat(context, *args, **kwargs):
    player = context['player']
    group = context['group']
    Constants = context['Constants']
    participant = context['participant']

    kwargs.setdefault('channel', group.id)
    kwargs.setdefault('nickname', 'Player {}'.format(player.id_in_group))

    # prefix the channel name with session code and app name
    context['channel'] = '{}-{}-{}'.format(
        context['session'].code,
        Constants.name_in_url,
        kwargs['channel']
    )

    nickname = kwargs['nickname']
    channel = context['channel']

    # channel name should not contain illegal chars,
    # so that it can be used in JS and URLs
    if not re.match(r'^[a-zA-Z0-9_-]+$', channel):
        raise ValueError(
            "'channel' can only contain ASCII letters, numbers, underscores, and hyphens. "
            "Value given was: {}".format(context['channel']))

    NicknameRegistration.objects.update_or_create(
        participant=participant,
        channel=channel,
        defaults={'nickname': nickname}
    )

    vars_for_js = {
        'channel': context['channel'],
        'participant_code': participant.code,
        'participant_id': participant.id
    }

    context['vars_for_js'] = safe_json(vars_for_js)

    return context

@register.filter(name='json_private')
def _json(value):
    if isinstance(value, QuerySet):
        value = list(value)
    return mark_safe(json.dumps(value))



from django.db import models
from otree.models import Participant, Session
import time


class ChatMessage(models.Model):
    class Meta:
        index_together = ['channel', 'timestamp']

    # the name "channel" here is unrelated to Django channels
    channel = models.CharField(max_length=255)
    participant = models.ForeignKey(Participant)
    nickname = models.CharField(max_length=255)

    # call it 'body' instead of 'message' or 'content' because those terms
    # are already used by channels
    body = models.TextField()
    timestamp = models.DateTimeField(auto_now_add=True)


class NicknameRegistration(models.Model):
    '''For security, have to register the nicknames on the server side'''

    class Meta:
        unique_together = ['channel', 'participant']

    channel = models.CharField(max_length=255)
    participant = models.ForeignKey(Participant)
    nickname = models.CharField(max_length=255)